@@ -1,21 +1,15 @@
 /*
  * Copyright 2003-2006 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
- * $Header: /var/cvsroot/gentoo-projects/pax-utils/scanelf.c,v 1.141 2006/04/23 15:24:38 flameeyes Exp $
+ * $Header: /var/cvsroot/gentoo-projects/pax-utils/scanelf.c,v 1.153 2006/05/17 21:45:20 solar Exp $
  *
  * Copyright 2003-2006 Ned Ludd        - <solar@gentoo.org>
  * Copyright 2004-2006 Mike Frysinger  - <vapier@gentoo.org>
  */
 
 #include "paxinc.h"
-#if defined(__GLIBC__) || defined(__UCLIBC__)
- #include <glob.h>
-#endif
-#if defined(__FreeBSD__) || defined(__DragonFly__)
- #include <elf-hints.h>
-#endif
 
-static const char *rcsid = "$Id: scanelf.c,v 1.141 2006/04/23 15:24:38 flameeyes Exp $";
+static const char *rcsid = "$Id: scanelf.c,v 1.153 2006/05/17 21:45:20 solar Exp $";
 #define argv0 "scanelf"
 
 #define IS_MODIFIER(c) (c == '%' || c == '#' || c == '+')
@@ -31,6 +25,7 @@ static const char *rcsid = "$Id: scanelf.c,v 1.141 2006/04/23 15:24:38 flameeyes
 
 
 /* prototypes */
+static int file_matches_list(const char *filename, char **matchlist);
 static int scanelf_elfobj(elfobj *elf);
 static int scanelf_elf(const char *filename, int fd, size_t len);
 static int scanelf_archive(const char *filename, int fd, size_t len);
@@ -39,9 +34,12 @@ static void scanelf_dir(const char *path);
 static void scanelf_ldpath(void);
 static void scanelf_envpath(void);
 static void usage(int status);
+static char **get_split_env(const char *envvar);
+static void parseenv(void);
 static void parseargs(int argc, char *argv[]);
 static char *xstrdup(const char *s);
 static void *xmalloc(size_t size);
+static void *xrealloc(void *ptr, size_t size);
 static void xstrncat(char **dst, const char *src, size_t *curr_len, size_t n);
 #define xstrcat(dst,src,curr_len) xstrncat(dst,src,curr_len,0)
 static inline void xchrcat(char **dst, const char append, size_t *curr_len);
@@ -78,11 +76,85 @@ static char fix_elf = 0;
 static char gmatch = 0;
 static char use_ldcache = 0;
 
+static char **qa_textrels = NULL;
+static char **qa_execstack = NULL;
+static char **qa_wx_load = NULL;
+
 int match_bits = 0;
 caddr_t ldcache = 0;
 size_t ldcache_size = 0;
 unsigned long setpax = 0UL;
 
+/* utility funcs */
+static char *xstrdup(const char *s)
+{
+	char *ret = strdup(s);
+	if (!ret) err("Could not strdup(): %s", strerror(errno));
+	return ret;
+}
+static void *xmalloc(size_t size)
+{
+	void *ret = malloc(size);
+	if (!ret) err("Could not malloc() %li bytes", (unsigned long)size);
+	return ret;
+}
+static void *xrealloc(void *ptr, size_t size)
+{
+	void *ret = realloc(ptr, size);
+	if (!ret) err("Could not realloc() %li bytes", (unsigned long)size);
+	return ret;
+}
+static void xstrncat(char **dst, const char *src, size_t *curr_len, size_t n)
+{
+	size_t new_len;
+
+	new_len = strlen(*dst) + strlen(src);
+	if (*curr_len <= new_len) {
+		*curr_len = new_len + (*curr_len / 2);
+		*dst = realloc(*dst, *curr_len);
+		if (!*dst)
+			err("could not realloc() %li bytes", (unsigned long)*curr_len);
+	}
+
+	if (n)
+		strncat(*dst, src, n);
+	else
+		strcat(*dst, src);
+}
+static inline void xchrcat(char **dst, const char append, size_t *curr_len)
+{
+	static char my_app[2];
+	my_app[0] = append;
+	my_app[1] = '\0';
+	xstrcat(dst, my_app, curr_len);
+}
+
+/* Match filename against entries in matchlist, return TRUE
+ * if the file is listed */
+static int file_matches_list(const char *filename, char **matchlist)
+{
+	char **file;
+	char *match;
+	char buf[__PAX_UTILS_PATH_MAX];
+
+	if (matchlist == NULL)
+		return 0;
+
+	for (file = matchlist; *file != NULL; file++) {
+		if (search_path) {
+			snprintf(buf, sizeof(buf), "%s%s", search_path, *file);
+			match = buf;
+		} else {
+			match = *file;
+		}
+		if (fnmatch(match, filename, 0) == 0)
+			return 1;
+	}
+	return 0;
+}
+
+
+
 /* sub-funcs for scanelf_file() */
 static void scanelf_file_get_symtabs(elfobj *elf, void **sym, void **tab)
 {
@@ -199,18 +271,25 @@ static char *scanelf_file_phdr(elfobj *elf, char *found_phdr, char *found_relro,
 		Elf ## B ## _Phdr *phdr = PHDR ## B (elf->phdr); \
 		for (i = 0; i < EGET(ehdr->e_phnum); ++i) { \
 			if (EGET(phdr[i].p_type) == PT_GNU_STACK) { \
-				if (multi_stack++) warnf("%s: multiple PT_GNU_STACK's !?", elf->filename); \
+				if (multi_stack++) \
+					warnf("%s: multiple PT_GNU_STACK's !?", elf->filename); \
+				if (file_matches_list(elf->filename, qa_execstack)) \
+					continue; \
 				found = found_phdr; \
 				offset = 0; \
 				check_flags = PF_X; \
 			} else if (EGET(phdr[i].p_type) == PT_GNU_RELRO) { \
-				if (multi_relro++) warnf("%s: multiple PT_GNU_RELRO's !?", elf->filename); \
+				if (multi_relro++) \
+					warnf("%s: multiple PT_GNU_RELRO's !?", elf->filename); \
 				found = found_relro; \
 				offset = 4; \
 				check_flags = PF_X; \
 			} else if (EGET(phdr[i].p_type) == PT_LOAD) { \
 				if (ehdr->e_type == ET_DYN || ehdr->e_type == ET_EXEC) \
-					if (multi_load++ > max_pt_load) warnf("%s: more than %i PT_LOAD's !?", elf->filename, max_pt_load); \
+					if (multi_load++ > max_pt_load) \
+						warnf("%s: more than %i PT_LOAD's !?", elf->filename, max_pt_load); \
+				if (file_matches_list(elf->filename, qa_wx_load)) \
+					continue; \
 				found = found_load; \
 				offset = 8; \
 				check_flags = PF_W|PF_X; \
@@ -271,6 +350,7 @@ static char *scanelf_file_phdr(elfobj *elf, char *found_phdr, char *found_relro,
 	else
 		return ret;
 }
+
 static const char *scanelf_file_textrel(elfobj *elf, char *found_textrel)
 {
 	static const char *ret = "TEXTREL";
@@ -278,6 +358,8 @@ static const char *scanelf_file_textrel(elfobj *elf, char *found_textrel)
 
 	if (!show_textrel && !show_textrels) return NULL;
 
+	if (file_matches_list(elf->filename, qa_textrels)) return NULL;
+
 	if (elf->phdr) {
 #define SHOW_TEXTREL(B) \
 	if (elf->elf_class == ELFCLASS ## B) { \
@@ -600,7 +682,9 @@ static void scanelf_file_rpath(elfobj *elf, char *found_rpath, char **ret, size_
 #define FLAG_MIPS64_LIBN64  0x0700
 
 static char *lookup_cache_lib(elfobj *, char *);
+
 #if defined(__GLIBC__) || defined(__UCLIBC__)
+
 static char *lookup_cache_lib(elfobj *elf, char *fname)
 {
 	int fd = 0;
@@ -666,8 +750,9 @@ static char *lookup_cache_lib(elfobj *elf, char *fname)
 	}
 	return buf;
 }
+
 #else
-#warning Cache support not implemented for your current target.
+#warning Cache support not implemented for your target
 static char *lookup_cache_lib(elfobj *elf, char *fname)
 {
 	return NULL;
@@ -904,7 +989,7 @@ static char *scanelf_file_sym(elfobj *elf, char *found_sym)
 						} \
 						/* ok, lets compare the name now */ \
 						if ((strncmp(this_sym, symname, (next_sym-this_sym)) == 0 && symname[next_sym-this_sym] == '\0') || \
-						    (strcmp(symname, versioned_symname) == 0)) { \
+						    (strncmp(symname, versioned_symname, strlen(versioned_symname)) == 0)) { \
 							if (be_semi_verbose) { \
 								char buf[126]; \
 								snprintf(buf, sizeof(buf), "%lX %s %s", \
@@ -1253,7 +1338,7 @@ static void scanelf_dir(const char *path)
 			      (unsigned long)len, (unsigned long)sizeof(buf));
 			continue;
 		}
-		sprintf(buf, "%s/%s", path, dentry->d_name);
+		snprintf(buf, sizeof(buf), "%s%s%s", path, (path[pathlen-1] == '/') ? "" : "/", dentry->d_name);
 		if (lstat(buf, &st) != -1) {
 			if (S_ISREG(st.st_mode))
 				scanelf_file(buf);
@@ -1289,13 +1374,14 @@ static int scanelf_from_file(const char *filename)
 }
 
 #if defined(__GLIBC__) || defined(__UCLIBC__)
-static int load_ld_so_conf(int i, const char *fname)
+
+static int load_ld_cache_config(int i, const char *fname)
 {
 	FILE *fp = NULL;
 	char *p;
 	char path[__PAX_UTILS_PATH_MAX];
 
-	if (i + 1 == sizeof(ldpaths) / sizeof(*ldpaths))
+	if (i + 1 == ARRAY_SIZE(ldpaths))
 		return i;
 
 	if ((fp = fopen(fname, "r")) == NULL)
@@ -1324,8 +1410,8 @@ static int load_ld_so_conf(int i, const char *fname)
 					/* try to avoid direct loops */
 					if (strcmp(gl.gl_pathv[x], fname) == 0)
 						continue;
-					i = load_ld_so_conf(i, gl.gl_pathv[x]);
-					if (i + 1 >= sizeof(ldpaths) / sizeof(*ldpaths)) {
+					i = load_ld_cache_config(i, gl.gl_pathv[x]);
+					if (i + 1 >= ARRAY_SIZE(ldpaths)) {
 						globfree64(&gl);
 						return i;
 					}
@@ -1340,7 +1426,7 @@ static int load_ld_so_conf(int i, const char *fname)
 
 		ldpaths[i++] = xstrdup(path);
 
-		if (i + 1 == sizeof(ldpaths) / sizeof(*ldpaths))
+		if (i + 1 == ARRAY_SIZE(ldpaths))
 			break;
 	}
 	ldpaths[i] = NULL;
@@ -1348,41 +1434,41 @@ static int load_ld_so_conf(int i, const char *fname)
 	fclose(fp);
 	return i;
 }
-#endif
 
-#if defined(__FreeBSD__) || (__DragonFly__)
-static int load_ld_so_hints(int i, const char *fname)
+#elif defined(__FreeBSD__) || (__DragonFly__)
+
+static int load_ld_cache_config(int i, const char *fname)
 {
 	FILE *fp = NULL;
 	char *b = NULL, *p;
 	struct elfhints_hdr hdr;
-	
-	if (i + 1 == sizeof(ldpaths) / sizeof(*ldpaths))
+
+	if (i + 1 == ARRAY_SIZE(ldpaths))
 		return i;
 
 	if ((fp = fopen(fname, "r")) == NULL)
 		return i;
 
-	if ( fread(&hdr, 1, sizeof(hdr), fp) != sizeof(hdr) ||
-		hdr.magic != ELFHINTS_MAGIC || hdr.version != 1 ||
-		fseek(fp, hdr.strtab + hdr.dirlist, SEEK_SET) == -1
-	) {
+	if (fread(&hdr, 1, sizeof(hdr), fp) != sizeof(hdr) ||
+	    hdr.magic != ELFHINTS_MAGIC || hdr.version != 1 ||
+	    fseek(fp, hdr.strtab + hdr.dirlist, SEEK_SET) == -1)
+	{
 		fclose(fp);
 		return i;
 	}
-	
+
 	b = (char*)malloc(hdr.dirlistlen+1);
-	if ( fread(b, 1, hdr.dirlistlen+1, fp) != hdr.dirlistlen+1 ) {
+	if (fread(b, 1, hdr.dirlistlen+1, fp) != hdr.dirlistlen+1) {
 		fclose(fp);
 		free(b);
 		return i;
 	}
-	
-	while ( (p = strsep(&b, ":")) ) {
-		if ( *p == '\0' ) continue;
+
+	while ((p = strsep(&b, ":"))) {
+		if (*p == '\0') continue;
 		ldpaths[i++] = xstrdup(p);
-		
-		if (i + 1 == sizeof(ldpaths) / sizeof(*ldpaths))
+
+		if (i + 1 == ARRAY_SIZE(ldpaths))
 			break;
 	}
 	ldpaths[i] = NULL;
@@ -1391,6 +1477,15 @@ static int load_ld_so_hints(int i, const char *fname)
 	fclose(fp);
 	return i;
 }
+
+#else
+
+#warning Cache config support not implemented for your target
+static int load_ld_cache_config(int i, const char *fname)
+{
+	memset(ldpaths, 0x00, sizeof(ldpaths));
+}
+
 #endif
 
 /* scan /etc/ld.so.conf for paths */
@@ -1574,10 +1669,8 @@ static void parseargs(int argc, char *argv[])
 			match_bits = atoi(optarg);
 			break;
 		case 'o': {
-			FILE *fp = NULL;
-			if ((fp = freopen(optarg, "w", stdout)) == NULL)
+			if (freopen(optarg, "w", stdout) == NULL)
 				err("Could not open output stream '%s': %s", optarg, strerror(errno));
-			SET_STDOUT(fp);
 			break;
 		}
 		case 'k':
@@ -1734,14 +1827,10 @@ static void parseargs(int argc, char *argv[])
 
 	/* now lets actually do the scanning */
 	if (scan_ldpath || use_ldcache)
-#if defined(__GLIBC__) || defined(__UCLIBC__)
-		load_ld_so_conf(0, "/etc/ld.so.conf");
-#elif defined(__FreeBSD__) || defined(__DragonFly__)
-		load_ld_so_hints(0, _PATH_ELF_HINTS);
-#endif
+		load_ld_cache_config(0, __PAX_UTILS_DEFAULT_LD_CACHE_CONFIG);
 	if (scan_ldpath) scanelf_ldpath();
 	if (scan_envpath) scanelf_envpath();
-	if (!from_file && optind == argc && ttyname(0) == NULL)
+	if (!from_file && optind == argc && ttyname(0) == NULL && !scan_ldpath && !scan_envpath)
 		from_file = "-";
 	if (from_file) {
 		scanelf_from_file(from_file);
@@ -1755,7 +1844,7 @@ static void parseargs(int argc, char *argv[])
 	}
 
 	/* clean up */
-	if (versioned_symname) free(versioned_symname);
+	free(versioned_symname);
 	for (i = 0; ldpaths[i]; ++i)
 		free(ldpaths[i]);
 
@@ -1763,45 +1852,54 @@ static void parseargs(int argc, char *argv[])
 		munmap(ldcache, ldcache_size);
 }
 
-
-
-/* utility funcs */
-static char *xstrdup(const char *s)
+static char **get_split_env(const char *envvar)
 {
-	char *ret = strdup(s);
-	if (!ret) err("Could not strdup(): %s", strerror(errno));
-	return ret;
-}
-static void *xmalloc(size_t size)
-{
-	void *ret = malloc(size);
-	if (!ret) err("Could not malloc() %li bytes", (unsigned long)size);
-	return ret;
-}
-static void xstrncat(char **dst, const char *src, size_t *curr_len, size_t n)
-{
-	size_t new_len;
+	const char *delims = " \t\n";
+	char **envvals = NULL;
+	char *env, *s;
+	int nentry;
 
-	new_len = strlen(*dst) + strlen(src);
-	if (*curr_len <= new_len) {
-		*curr_len = new_len + (*curr_len / 2);
-		*dst = realloc(*dst, *curr_len);
-		if (!*dst)
-			err("could not realloc() %li bytes", (unsigned long)*curr_len);
+	if ((env = getenv(envvar)) == NULL)
+		return NULL;
+
+	env = xstrdup(env);
+	if (env == NULL)
+		return NULL;
+
+	s = strtok(env, delims);
+	if (s == NULL) {
+		free(env);
+		return NULL;
 	}
 
-	if (n)
-		strncat(*dst, src, n);
-	else
-		strcat(*dst, src);
+	nentry = 0;
+	while (s != NULL) {
+		++nentry;
+		envvals = xrealloc(envvals, sizeof(*envvals) * (nentry+1));
+		envvals[nentry-1] = s;
+		s = strtok(NULL, delims);
+	}
+	envvals[nentry] = NULL;
+
+	/* don't want to free(env) as it contains the memory that backs
+	 * the envvals array of strings */
+	return envvals;
 }
-static inline void xchrcat(char **dst, const char append, size_t *curr_len)
+static void parseenv()
 {
-	static char my_app[2];
-	my_app[0] = append;
-	my_app[1] = '\0';
-	xstrcat(dst, my_app, curr_len);
+	qa_textrels = get_split_env("QA_TEXTRELS");
+	qa_execstack = get_split_env("QA_EXECSTACK");
+	qa_wx_load = get_split_env("QA_WX_LOAD");
 }
+#ifdef __PAX_UTILS_CLEANUP
+static void cleanup()
+{
+	free(out_format);
+	free(qa_textrels);
+	free(qa_execstack);
+	free(qa_wx_load);
+}
+#endif
 
 
 
@@ -1809,10 +1907,14 @@ int main(int argc, char *argv[])
 {
 	if (argc < 2)
 		usage(EXIT_FAILURE);
+	parseenv();
 	parseargs(argc, argv);
 	fclose(stdout);
-#ifdef __BOUNDS_CHECKING_ON
-	warn("The calls to add/delete heap should be off by 1 due to the out_buffer not being freed in scanelf_file()");
+#ifdef __PAX_UTILS_CLEANUP
+	cleanup();
+	warn("The calls to add/delete heap should be off:\n"
+	     "\t- 1 due to the out_buffer not being freed in scanelf_file()\n"
+	     "\t- 1 per QA_TEXTRELS/QA_EXECSTACK/QA_WX_LOAD");
 #endif
 	return EXIT_SUCCESS;
 }
